#include <cassert>
#include <chrono>
#include <iostream>

#include "singly-linked-list.h"

namespace
{
   /*
    * run a simple test
    */
   template<typename SeqnT, typename ElmnT>
      void test_sequence(SeqnT* sqn, int n, int m, ElmnT a, ElmnT b, std::ostream* os)
   {
      assert((m > 0) && (n > m));
      if(os) *os << "Push " << a << "'s.\n";
      for(int i = 0; i < n; i++) sqn->push(a);
      
      if(os) *os << "Overwrite element at index " << m/2 << " with " << b << ".\n";
      ElmnT& item = sqn->at(m/2);
      item = b;
      
      if(os) *os << "\nSize of sequence: " << sqn->size() << ".\n";
      if(os) *os << "Detaching " << m/2 << " elements from front:";
      for(int i = 0; i < m/2; i++)
      {
         if(os) *os << " " << sqn->front();
         sqn->pop_front();
      }
      if(os) *os << ".\nDetaching " << m/2 << " elements from back:";
      for(int i = 0; i < m/2; i++)
      {
         if(os) *os << " " << sqn->back();
         sqn->pop_back();
      }
      if(os) *os << ".\nSize of sequence: " << sqn->size() << ".\n";
      
      if(os) *os << "\nElement at index " << (n-m)/3 << ": " << sqn->at((n-m)/3) << ".\n";
      if(os) *os << "Element at index " << 2*(n-m)/3 << ": " << sqn->at(2*(n-m)/3) << ".\n";
      
      if(os) *os << "\nInsert " << b << " at index " << 2*(n-m)/3 << ".\n";
      sqn->insert_at(2*(n-m)/3, b);
      if(os) *os << "Size of sequence: " << sqn->size() << ".\n";
      
      if(os) *os << "\nDelete element at index " << (n-m)/3 << ".\n";
      sqn->erase_at((n-m)/3);
      if(os) *os << "Size of sequence: " << sqn->size() << ".\n";
      if(os) *os << "Element at index " << 2*(n-m)/3 - 1 << ": " << sqn->at(2*(n-m)/3 - 1) << ".\n";
      
      if(os) *os << "\nClearing.\n";
      sqn->clear();
      if(os) *os << "Size of sequence: " << sqn->size() << ".\n";
   }
   
   /*
    * return time measurement in units of seconds
    */
   template<typename SeqnT, typename ElmnT>
      float test_with_time_measurement(SeqnT* sqn, int iterations, ElmnT a, ElmnT b)
   {
      int sequence_length = 1000001;
      int deletions = 10;
      test_sequence(sqn, sequence_length, deletions, a, b, &std::cout);

      int log_entries = 10;
      std::cout << "\nNow repeat the above " << iterations << " times:\n";
      auto t0 = std::chrono::high_resolution_clock::now();
      for(int i = 0; i < iterations; i++)
      {
         test_sequence(sqn, sequence_length, deletions, a, b, nullptr);
         if((i+1) % (iterations/log_entries) == 0)
         {
            std::cout << "\t" << i+1 << "\n";
            std::cout.flush();  // make sure that status output is shown without delay
         }
      }
      auto t1 = std::chrono::high_resolution_clock::now();
      return 1.0e-06 * std::chrono::duration_cast<std::chrono::microseconds>(t1-t0).count();
   }
}

int main()
{
   int iterations = 250;
   
   std::cout << "\n\n*** test with singly linked list of integers ***\n";
   seq::SinglyLinkedList<int> sll_int;
   float sll_int_time = test_with_time_measurement(&sll_int, iterations, 4, 2);
   
   std::cout << "\n\n*** test with singly linked list of strings ***\n";
   seq::SinglyLinkedList<std::string> sll_string;
   std::string string_a = "four";
   std::string string_b = "two";
   float sll_string_time = test_with_time_measurement(&sll_string, iterations, string_a, string_b);
   
   std::cout << "\n\nRuntime for SinglyLinkedList<int>:\t" << sll_int_time << " s\n";
   std::cout << "Runtime for SinglyLinkedList<string>:\t" << sll_string_time << " s\n";
}
