
#include <iostream>
#include "singly-linked-list.h"

using namespace seq;

template class seq::SinglyLinkedList<int>;
template class seq::SinglyLinkedList<std::string>;

// return the size (number of items in the singly linked list)
template<typename T> size_t SinglyLinkedList<T>::size() const
{
   size_t count = 0;
   for(auto n = this->head; n != nullptr; n = n->get_next()) count++;
   return count;
}

// return pointer to the node at position i, counting from 0
template<typename T> SinglyLinkedListNode<T>* SinglyLinkedList<T>::index(int i) const
{
   auto n = this->head;
   for(int k = 0; k < i; k++)
   {
      assert(n != nullptr);
      n = n->get_next();
   }
   return n;
}

// add an item at the end of the list
template<typename T> void SinglyLinkedList<T>::push_back(const T& pushed_item)
{
   SinglyLinkedListNode<T>* new_node = new SinglyLinkedListNode<T>;
   new_node->set_item(pushed_item);
   
   if(this->empty()) this->head = new_node;
   else this->tail->set_next(new_node);
   this->tail = new_node;
}

// add an item at the beginning of the list
template<typename T> void SinglyLinkedList<T>::push_front(const T& pushed_item)
{
   SinglyLinkedListNode<T>* new_node = new SinglyLinkedListNode<T>;
   new_node->set_item(pushed_item);
   
   if(this->empty()) this->tail = new_node;
   else new_node->set_next(this->head);  /* there was a bug here, thanks to Hanna Lye Moum and Nora Mikarlsen for fixing */
   this->head = new_node;
}

// remove the head node and item
template<typename T> void SinglyLinkedList<T>::pop_front()
{
   if(this->empty()) return;  // nothing there to remove
   auto successor = this->head->get_next();
   
   delete this->head;
   this->head = successor;  // successor of the previous head is the new head
   if(this->head == nullptr) this->tail = nullptr;  // catch special case: the list is now empty
}

// remove the tail node and item
// note: in a singly linked list, we need to walk step by step to the end!
template<typename T> void SinglyLinkedList<T>::pop_back()
{
   if(this->empty()) return;  // nothing there to remove
   
   // list only contains one element
   if(this->head->get_next() == nullptr)
   {
      delete this->head;
      this->head = nullptr;
      this->tail = nullptr;
      return;
   }
   
   /*
    * walk through until "third" is nullptr, second is the last element, and first is the second-to-last
    */
   auto first = this->head;
   auto second = first->get_next();
   auto third = second->get_next();
   while(third != nullptr)
   {
      first = second;
      second = third;
      third = third->get_next();
   }
   
   /*
    * remove "second"
    */
   if(this->tail == second) this->tail = first;
   first->set_next(nullptr);
   delete second;
}

// insert an item at index i
template<typename T> void SinglyLinkedList<T>::insert_at(int i, const T& inserted_item)
{
   if(i == 0)
   {
      this->push_front(inserted_item);
      return;
   }
   auto predecessor = this->index(i-1);
   this->insert_successor_to(predecessor, inserted_item);
}

// remove the item at index i
template<typename T> void SinglyLinkedList<T>::erase_at(int i)
{
   if(i == 0)
   {
      this->pop_front();
      return;
   }
   auto predecessor = this->index(i-1);
   this->erase_successor_to(predecessor);
}

// insert an item after given node
template<typename T> void SinglyLinkedList<T>::insert_successor_to(SinglyLinkedListNode<T>* predecessor, const T& inserted_item)
{
   SinglyLinkedListNode<T>* new_node =  new SinglyLinkedListNode<T>;
   new_node->set_item(inserted_item);
   
   auto successor = predecessor->get_next();
   predecessor->set_next(new_node);
   new_node->set_next(successor);
   
   if(!successor) this->tail = new_node;
}

// remove the item after given node
template<typename T> void SinglyLinkedList<T>::erase_successor_to(SinglyLinkedListNode<T>* predecessor)
{
   auto erased_node = predecessor->get_next();
   auto new_successor = erased_node->get_next();
   
   predecessor->set_next(new_successor);
   if(new_successor == nullptr) this->tail = predecessor;
   delete erased_node;
}
